// ======================================================================
//
// Extents.cpp
// jeff grills
//
// copyright 1997 origin systems inc
//
// ======================================================================

#include "First.h"
#include "Extents.h"

#include "Gl.h"
#include "Iff.h"

// ======================================================================

#define DPRINT_F(a) NOP

// ======================================================================

Extents::Node::Node(Iff &iff)
{
	iff.seekChunk(TAG_DATA);
	type = static_cast<Type>(iff.read_int32());
	iff.read_floatVector(1, &normal);
	planeD = iff.read_float();
	region = iff.read_int32();
	iff.exitChunk();

	vertex = NULL;
	vertexCount = 0;

	if (type == split)
	{
		iff.seekForm(TAG_BACK);
		back = NEW Node(iff);
		iff.exitForm();

		iff.seekForm(TAG_FRNT);
		front = NEW Node(iff);
		iff.exitForm();

		if (iff.seekChunk(TAG_VERT, Iff::FlagOptional))
		{
			vertexCount = iff.chunkLength(sizeof(Vector::Save));
			vertex = iff.read_floatVector(Iff::ReadRest);
			iff.exitChunk();
		}
	}
	else
	{
		front = NULL;
		back = NULL;
	}
}

// ----------------------------------------------------------------------

Extents::Node::~Node(void)
{
	if (type == split)
	{
		delete back;
		delete front;
	}

	back = NULL;
	front = NULL;

	delete [] vertex;
}

// ----------------------------------------------------------------------

void Extents::Node::render(void) const
{	
	int     i, j;

	if (type == split)
	{
		// draw the polygon edges
		Gl::setMaterial(MaterialList::blue());

		for (j = vertexCount-1, i = 0; i < vertexCount; j = i++)
			Gl::line(vertex[j], vertex[i]);

		back->render();
		front->render();
	}
}

// ----------------------------------------------------------------------

bool Extents::Node::collide(const Vector &point, const Node **node, const Node **splitNode) const
{
	bool result;
	real side;

	switch (type)
	{
		case split:
	
			side = normal.dot(point) + planeD;
			*splitNode = this;

			if (side < CONST_REAL(0))
				result = back->collide(point, node, splitNode);
			else
				if (side > CONST_REAL(0))
					result = front->collide(point, node, splitNode);
				else
					result = false;
			break;

		case full:
			*node = this;
			result = true;
			break;

		case empty:
			result = false;
			break;

		default:
			result = false;
			DEBUG_FATAL(true, ("Extents::Node::collide bad type"));
	}

	return result;
}

// ----------------------------------------------------------------------

bool Extents::Node::collide(const Vector &start, const Vector &stop, Vector *intersectLocation, const Node **node, const Node **splitNode) const
{
	real   startSide, stopSide, t;
	Vector intersect;

	switch (type)
	{
		case split:

			startSide = normal.dot(start) + planeD;
			stopSide  = normal.dot(stop)  + planeD;

			if (opposingSignsNotZero(startSide, stopSide))
			{
				*splitNode = this;

				// compute the intersection point
				t = abs(startSide) / (abs(startSide) + abs(stopSide));
				DEBUG_FATAL(t < 0 || t > 1, ("Extents::Node::collide bad t %f", t));
				intersect.sub(stop, start);
				intersect *= t;
				intersect += start;

				if (startSide < CONST_REAL(0))
					if (back->collide(start, intersect, intersectLocation, node, splitNode))
						return true;
					else
						return front->collide(intersect, stop, intersectLocation, node, splitNode);
				else
					if (front->collide(start, intersect, intersectLocation, node, splitNode))
						return true;
					else
						return back->collide(intersect, stop, intersectLocation, node, splitNode);
			}
			else
				if (startSide < CONST_REAL(0) || stopSide < CONST_REAL(0))
					return back->collide(start, stop, intersectLocation, node, splitNode);
				else
					if (startSide > CONST_REAL(0) || stopSide > CONST_REAL(0))
						return front->collide(start, stop, intersectLocation, node, splitNode);
					else
					{
						// line is contained within the splitting plane.
						// -q- not much of an intersection, so just return false
						return false;
					}

		case full:
			if (intersectLocation)
				*intersectLocation = start;
			*node = this;
			return true;

		case empty:
			return false;
	}
	
	DEBUG_FATAL(true, ("Extents::Node::collide bad type"));
	return false; //lint !e527 // unreachable
}

// ======================================================================

Extents::Extents(const Vector &newCenter, real newRadius)
{
	init();
	sphereCenter = newCenter;
	sphereRadius = newRadius;
}

// ----------------------------------------------------------------------

Extents::Extents(Iff &iff)
{
	int i;

	init();

	if (iff.seekChunk(TAG_SPHR, Iff::FlagOptional))
	{
		iff.read_floatVector(1, &sphereCenter);
		sphereRadius = iff.read_float();
		iff.exitChunk();
	}

	if (iff.seekForm(TAG_EXTN, Iff::FlagOptional))
	{	
		// load the extent tree
		iff.seekForm(TAG_TREE);
		root = NEW Node(iff);
		iff.exitForm();

		// load the region display stuff
		iff.seekForm(TAG_REGN);
		regions = iff.blockCount();
		regionVertex = NEW Vector*[regions];
		regionVertexCount = NEW int8[regions];
		regionVertexIndex = NEW int8*[regions];
		for (i = 0; i < regions; i++)
		{
			iff.seekForm(TAG_BLCK);

			iff.seekChunk(TAG_VERT);
			regionVertexCount[i] = static_cast<int8>(iff.chunkLength(sizeof(Vector::Save)));
			regionVertex[i] = MEM_OWN(iff.read_floatVector(Iff::ReadRest));
			iff.exitChunk();

			iff.seekChunk(TAG_INDX);
			regionVertexIndex[i] = MEM_OWN(iff.read_int8(Iff::ReadRest));
			iff.exitChunk();
		
			iff.exitForm();
		}
		iff.exitForm();
		iff.exitForm();
	}
}

// ----------------------------------------------------------------------

Extents::~Extents(void)
{
	int i;

	for (i = 0; i < regions; i++)
	{
		delete [] regionVertex[i];
		delete [] regionVertexIndex[i];
	}
	delete [] regionVertex;
	delete [] regionVertexCount;
	delete [] regionVertexIndex;

	delete root;
}

// ----------------------------------------------------------------------

void Extents::init(void)
{
	users = 0;

	sphereCenter.makeZero();
	sphereRadius = CONST_REAL(0);

	root = NULL;
	regions = 0;
	regionVertex = NULL;
	regionVertexCount = NULL;
	regionVertexIndex = NULL;
}

// ----------------------------------------------------------------------

bool Extents::collideSpheres(const Vector &center, real radius) const
{
	radius += sphereRadius;

	return (center.magnitudeSquared() < (radius * radius));
}

// ----------------------------------------------------------------------

bool Extents::collideSpheres(const Vector &start, const Vector &end, real radius, Vector *result) const
{
	Vector v;
	real   v_dot_v;
	bool   ret_val = false;

	v.sub(end, start);
	v_dot_v = v.dot(v);

	if(collideSpheres(start, radius))
	{
		if(result)
			*result = start;

		ret_val = true;
	}
	else if(v_dot_v > CONST_REAL(0))
	{
		Vector x;
		real   x_dot_x, x_dot_v, time_closest;

		x = start;
		x_dot_x = x.dot(x);
		x_dot_v = x.dot(v);

		time_closest = -x_dot_v / v_dot_v;

		if(time_closest < CONST_REAL(0))
		{
			// uniform divergence
			// a collision could not have occurred since we already know
			// that we weren't colliding at the start point
			ret_val = false;
		}
		else if(time_closest >= CONST_REAL(1))
		{
			real closest_squared, touch_squared;
			
			// uniform convergence
			// check the end point for a collision
			closest_squared = end.dot(end);
			touch_squared   = (sphereRadius + radius) * (sphereRadius + radius);

			if(closest_squared < touch_squared)
			{
				if(result)
				{
					time_closest = (x_dot_v * x_dot_v) - (v_dot_v * (x_dot_x - touch_squared));
					DEBUG_FATAL(time_closest < CONST_REAL(0), ("Extent::collideSpheres negative number under radical"));
					time_closest = -sqrt(time_closest);
					time_closest -= x_dot_v;
					time_closest /= v_dot_v;

					DEBUG_FATAL(time_closest < CONST_REAL(0), ("Extent::collideSpheres time_closest < 0"));
					DEBUG_FATAL(time_closest > CONST_REAL(1), ("Extent::collideSpheres time_closest > 1"));

					*result  = v;
					*result *= time_closest;
					*result += start;
				}

				ret_val = true;
			}
		}
		else
		{
			Vector temp;
			real   closest_squared, touch_squared;

			// converge then diverge
			temp  = v;
			temp *= time_closest;
			temp += start;

			closest_squared = temp.dot(temp);
			touch_squared   = (sphereRadius + radius) * (sphereRadius + radius);

			if(closest_squared < touch_squared)
			{
				if(result)
				{
					time_closest = (x_dot_v * x_dot_v) - (v_dot_v * (x_dot_x - touch_squared));
					DEBUG_FATAL(time_closest < CONST_REAL(0), ("Extent::collideSpheres negative number under radical"));
					time_closest = -sqrt(time_closest);
					time_closest -= x_dot_v;
					time_closest /= v_dot_v;

					DEBUG_FATAL(time_closest < CONST_REAL(0), ("Extent::collideSpheres time_closest < 0"));
					DEBUG_FATAL(time_closest > CONST_REAL(1), ("Extent::collideSpheres time_closest > 1"));

					*result  = v;
					*result *= time_closest;
					*result += start;
				}

				ret_val = true;
			}
		}
	}

	return ret_val;
}

// ----------------------------------------------------------------------

bool Extents::collide(const Vector &point, int *regionNumber, Vector *normal) const
{
	const Node *node, *split;

	node  = NULL;
	split = NULL;

	if (root && root->collide(point, &node, &split))
	{
		if (regionNumber)
			*regionNumber = node->getRegion();

		if (normal)
		{
			if (split)
				*normal = split->getNormal();
			else
				*normal = Vector::unitY;
		}

		return true;
	}

	return false;
}

// ----------------------------------------------------------------------

bool Extents::collide(const Vector &start, const Vector &end, Vector *result, int *regionNumber, Vector *normal) const
{
	const Node *node, *split;

	node  = NULL;
	split = NULL;

	if (root && root->collide(start, end, result, &node, &split))
	{
		if (regionNumber)
			*regionNumber = node->getRegion();

		if (normal)
		{
			if(split)
				*normal = split->getNormal();
			else
				*normal = Vector::zero;
		}
		return true;
	}

	return false;
}

// ----------------------------------------------------------------------

void Extents::render(void) const
{	
	if (root)
		root->render();
}

// ======================================================================
